"""
FILE    : ION_Connect.py
AUTHOR  : Talha.Ali
VERSION : 1.0.0
DATE MODIFIED: 13.11.2024
PURPOSE : ION Connect Protocol Implementation
"""

from serial import Serial 
import minimalmodbus
import smbus2 as SMBus 

""" Register Address List """

# Board Information (Non-volatile)
BOARD_SERIAL_NUMBER = 0x0000  # Serial number of the board
BOARD_BATCH_NUMBER  = 0x0002  # Batch number of the board
PRODUCTION_DATE = 0x0004  # Production date of the board
PCB_HARDWARE_VERSION = 0x0006  # Hardware version of the PCB
FIRMWARE_VERSION = 0x0007  # Firmware version

# Board Configuration (Non-volatile)
SENSOR_TYPE = 0x000B  # Type of sensor
RESPONSE_FACTOR_VALUE = 0x000C  # Response factor value for the sensor
SENSOR_MAXIMUM_VALUE = 0x000E  # Maximum value the sensor can read
USE_SENSOR_MAXIMUM_VALUE = 0x0010  # Flag to use the sensor's maximum value
ALARM_THRESHOLD_VALUE = 0x0011  # Threshold value for triggering an alarm
ALARM_LATCH = 0x0013  # Alarm latch status
LED_CONTROL = 0x0014  # Control for LED indicators
MODBUS_SLAVE_ADDRESS = 0x0015  # Modbus slave address
UART_MODBUS_BAUD_RATE = 0x0016  # Baud rate for UART Modbus communication
I2C_SLAVE_ADDRESS = 0x0017  # I2C slave address

# Calibration Data (Non-volatile)
CALIBRATION_DATE = 0x001C  # Date of the last calibration
CALIBRATION_RESPONSE_FACTOR = 0x001E  # Response factor used during calibration
ZERO_RESPONSE = 0x0020  # Zero response value
SPAN_1_CONCENTRATION = 0x0022  # Concentration for span 1
SPAN_1_RESPONSE = 0x0024  # Response for span 1
SPAN_2_CONCENTRATION = 0x0026  # Concentration for span 2
SPAN_2_RESPONSE = 0x0028  # Response for span 2
SPAN_3_CONCENTRATION = 0x002A  # Concentration for span 3
SPAN_3_RESPONSE = 0x002C  # Response for span 3
SPAN_4_CONCENTRATION = 0x002E  # Concentration for span 4
SPAN_4_RESPONSE = 0x0030  # Response for span 4

# Sensing Telemetry (Volatile)
SENSOR_MV_VALUE = 0x0034  # Sensor millivolt value
SENSOR_PPM_VALUE = 0x0036  # Sensor parts per million value
TEMPERATURE = 0x0038  # Temperature reading
HUMIDITY = 0x003A  # Humidity reading
PRESSURE = 0x003C  # Pressure reading
CURRENT_LOOP_4MA = 0x003E  # Current loop 4mA value
CURRENT_LOOP_20MA = 0x0040  # Current loop 20mA value
CURRENT_LOOP_4_20MA_SENSITIVITY = 0x0042  # Sensitivity of the 4-20mA current loop
DAC_VALUE = 0x0044  # Digital-to-Analog Converter value
DAC_DATA = 0x0046  # Data for the DAC

# Status Registers (Volatile)
WORKING_STATUS = 0x0047  # Working status of the device
SENSOR_ERROR_STATUS = 0x0048  # Error status of the sensor

# Control Registers (Volatile)
PID_SUPPLY = 0x0050  # PID supply value
ALARM_LATCH_CLEAR = 0x0051  # Clear the alarm latch
OPERATION_MODE = 0x0052  # Operation mode of the device
NEW_RELATIVE_ZERO = 0x0054  # New relative zero value
USE_NEW_RELATIVE_ZERO = 0x0056  # Flag to use the new relative zero value

# System Commands (Volatile)
SOFTWARE_RESET = 0x0058  # Command to reset the software

# Default Addresses
DEFAULT_MODBUS_ADDRESS = 0x0001  # Default Modbus address
DEFAULT_I2C_ADDRESS = 0x0015  # Default I2C address

class Modbus:
    """Base functionality for Modbus Protocol

    **Quickstart: Importing and using Ion_Connect Modbus Protocol**

        Here is an example of using the :class:`Modbus`.
        First you will need to import the libraries to use the sensor

        .. code-block:: python

            import ION_Connect

        Once this is done you can define your `Ion_Connect.Modbus` object and define the object

        .. code-block:: python

            port = '/dev/ttyUSB0'
            baudrate = 9600
            timeout = 1
            ion_modbus = Ion_Connect.Modbus(port, baudrate, timeout)

        Now you have access to the registers.
        The following example will read the sensor mV and return a float to :variable:'sensor_mv'.

        .. code-block:: python

            sensor_mv = ion_modbus.get_sensor_mv()

        The following example will set the response factor to value 2 and return a float to :variable:'response_factor'.
        In order to set the response factor value, the operation mode requires to be set to 1 in order to access configure mode.

        .. code-block:: python

            ion_modbus.set_operation_mode(1)
            ion_modbus.set_response_factor(2)
            ion_modbus.set_operation_mode(0)

            response_factor = ion_modbus.get_response_factor()

    """
    def __init__(self, port: str, baudrate: int = 9600, timeout: int = 1, parity: str = 'E', stopbits: int = 1, bytesize: int = 8) -> None:
        """
        Initializes the Modbus communication protocol.

        :param port: Port to query from.
        :type port: str
        :param baudrate: Baud rate used for communication. Default is :const:'9600'.
        :type baudrate: int
        :param timeout: Timeout duration after queries are sent. Default is 1 second.
        :type timeout: int
        """
        if not port:
            raise ConnectionError("No port has been defined!")
        self._port = port
        self._baudrate = baudrate
        self._timeout = timeout
        self._parity = parity
        self._stopbits = stopbits
        self._bytesize = bytesize
        self._modbus = None
    @property
    def port(self) -> str:
        """
        Gets the port.
                    
        :returns: port value.
        :rtype: str
        """
        return self._port
    @property
    def baudrate(self) -> int:
        """
        Gets the baudrate            
            
        :returns: baudrate value.
        :rtype: int
        """
        return self._baudrate
    @property
    def timeout(self) -> int:
        """
        Gets the timeout.
            
        :returns: timeout value.
        :rtype: int
        """
        return self._timeout
    @property
    def parity(self) -> str:
        """
        Gets the parity.
            
        :returns: parity value.
        :rtype: int
        """
        return self._parity
    @property
    def stopbits(self) -> int:
        """
        Gets the stopbits.
            
        :returns: stopbits value.
        :rtype: int
        """
        return self._stopbits
    @property
    def bytesize(self) -> int:
        """
        Gets the bytesize.
            
        :returns: bytesize value.
        :rtype: int
        """
        return self._bytesize
    def open(self) -> None:
        """Opens the Modbus connection."""
        if self._modbus is None:
            try:
                self._modbus = minimalmodbus.Instrument(self._port, minimalmodbus.MODE_RTU)
                self._modbus.serial.bytesize = self._bytesize
                self._modbus.serial.baudrate = self._baudrate
                self._modbus.serial.parity = self._parity
                self._modbus.serial.stopbits = self._stopbits
                self._modbus.serial.timeout = self._timeout
                self._modbus.clear_buffers_before_each_transaction = True
                self._modbus.close_port_after_each_call = True
                self.flush()
            except Exception as e:
                raise ConnectionError(f"Failed to open Modbus connection: {e}")
    def close(self) -> None:
        """Closes the Modbus connection."""
        if self._modbus is not None:
            try:
                self._modbus.serial.close()
            except Exception as e:
                print(f"Failed to close Modbus connection: {e}")
            finally:
                self._modbus = None
    def __del__(self) -> None:
        """Ensures the Modbus connection is closed when the object is deleted."""
        self.close()
    def _set_address(self, slave_address: int) -> None:
        """
        Sets and validates the Modbus address.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :raises ValueError: If the slave address exceeds 247.
        """
        if slave_address > 247:
            raise ValueError("Slave Address cannot exceed 247!")
        self._modbus.address = slave_address
    def flush(self) -> None:
        """Flushes the Modbus connection."""
        self._modbus.serial.reset_input_buffer()
        self._modbus.serial.reset_output_buffer()
    def get_board_serial_number(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Board Serial Number from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Board serial number as a 32-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_long(registeraddress=BOARD_SERIAL_NUMBER)
        except Exception as e:
            print(f"Failed to read board serial number: {e}")
            return -1
    def get_board_batch_number(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Board Batch Number from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Board batch number as a 32-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_long(registeraddress=BOARD_BATCH_NUMBER)
        except Exception as e:
            print(f"Failed to read board batch number: {e}")
            return -1
    def get_production_date(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Production Date from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Production date as a 32-bit unsigned integer value in format YYYYMMDD.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_long(registeraddress=PRODUCTION_DATE)
        except Exception as e:
            print(f"Failed to read production date: {e}")
            return -1
    def get_pcb_hardware(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the PCB Hardware version from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: PCB hardware version as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=PCB_HARDWARE_VERSION)
        except Exception as e:
            print(f"Failed to read PCB hardware version: {e}")
            return -1
    def get_firmware_version(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Firmware Version from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Firmware version as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=FIRMWARE_VERSION)
        except Exception as e:
            print(f"Failed to read firmware version: {e}")
            return -1
    def get_sensor_type(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Sensor Type from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Sensor type as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=SENSOR_TYPE)
        except Exception as e:
            print(f"Failed to read sensor type: {e}")
            return -1
    def set_sensor_type(self, sensor_type: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Sensor Type to registers.

        :param sensor_type: Sensor type as a 16-bit unsigned integer value.
        :type sensor_type: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=SENSOR_TYPE, value=sensor_type)
        except Exception as e:
            print(f"Failed to write sensor type: {e}")
    def get_response_factor(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Response Factor from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Response factor as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=RESPONSE_FACTOR_VALUE)
        except Exception as e:
            print(f"Failed to read response factor: {e}")
            return -1.0
    def set_response_factor(self, response_factor: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Response Factor to registers.

        :param response_factor: Response factor as a 32-bit float value.
        :type response_factor: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=RESPONSE_FACTOR_VALUE, value=response_factor)
        except Exception as e:
            print(f"Failed to write response factor: {e}")
    def get_sensor_maximum_value(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Sensor Maximum Value from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Sensor maximum value as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=SENSOR_MAXIMUM_VALUE)
        except Exception as e:
            print(f"Failed to read sensor maximum value: {e}")
            return -1.0
    def set_sensor_maximum_value(self, sensor_maximum: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Sensor Maximum Value to registers.

        :param sensor_maximum: Sensor maximum value as a 32-bit float value.
        :type sensor_maximum: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=SENSOR_MAXIMUM_VALUE, value=sensor_maximum)
        except Exception as e:
            print(f"Failed to write sensor maximum value: {e}")
    def get_sensor_maximum_state(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Sensor Maximum State from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Sensor maximum state as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=USE_SENSOR_MAXIMUM_VALUE)
        except Exception as e:
            print(f"Failed to read sensor maximum state: {e}")
            return -1
    def set_sensor_maximum_state(self, sensor_maximum_state: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Sensor Maximum State to registers.

        :param sensor_maximum_state: Sensor maximum state as a 16-bit unsigned integer value.
        :type sensor_maxmimum_state: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=USE_SENSOR_MAXIMUM_VALUE, value=sensor_maximum_state)
        except Exception as e:
            print(f"Failed to write sensor maximum state: {e}")
    def get_alarm_threshold_value(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Alarm Threshold Value from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Alarm threshold value as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=ALARM_THRESHOLD_VALUE)
        except Exception as e:
            print(f"Failed to read alarm threshold value: {e}")
            return -1.0
    def set_alarm_threshold_value(self, alarm_threshold_value: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Alarm Threshold Value to registers.

        :param alarm_threshold_value: Alarm threshold value as a 32-bit float value.
        :type alarm_threshold_value: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=ALARM_THRESHOLD_VALUE, value=alarm_threshold_value)
        except Exception as e:
            print(f"Failed to write alarm threshold value: {e}")
    def get_alarm_latch_state(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Alarm Latch state from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Alarm latch as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=ALARM_LATCH)
        except Exception as e:
            print(f"Failed to read alarm latch state: {e}")
            return -1
    def set_alarm_latch_state(self, alarm_latch_state: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Alarm Latch State to registers.

        :param alarm_latch_state: Alarm latch state as a 16-bit unsigned integer value.
        :type alarm_latch_state: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=ALARM_LATCH, value=alarm_latch_state)
        except Exception as e:
            print(f"Failed to write alarm latch state: {e}")
    def get_led_control(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the LED Control from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: LED control as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=LED_CONTROL)
        except Exception as e:
            print(f"Failed to read LED control: {e}")
            return -1
    def set_led_control(self, led_control: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the LED Control Status to registers.

        :param led_control: LED control state as a 16-bit unsigned integer value.
        :type led_control: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=LED_CONTROL, value=led_control)
        except Exception as e:
            print(f"Failed to write LED control: {e}")
    def get_modbus_slave_address(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Modbus Slave Address from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Modbus slave address as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=MODBUS_SLAVE_ADDRESS)
        except Exception as e:
            print(f"Failed to read Modbus slave address: {e}")
            return -1
    def set_modbus_slave_address(self, modbus_slave_address: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Modbus Slave Address to registers.

        :param modbus_slave_address: Modbus slave address as a 16-bit unsigned integer value.
        :type modbus_slave_address: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=MODBUS_SLAVE_ADDRESS, value=modbus_slave_address)
        except Exception as e:
            print(f"Failed to write Modbus slave address: {e}")
    def get_uart_modbus_baudrate(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the UART Modbus Baudrate from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: UART Modbus baudrate as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=UART_MODBUS_BAUD_RATE)
        except Exception as e:
            print(f"Failed to read UART Modbus baudrate: {e}")
            return -1
    def set_uart_modbus_baudrate(self, uart_modbus_baudrate: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the UART Modbus Baudrate to registers.

        :param uart_modbus_baudrate: UART Modbus baudrate as a 16-bit unsigned integer value.
        :type uart_modbus_baudrate: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=UART_MODBUS_BAUD_RATE, value=uart_modbus_baudrate)
        except Exception as e:
            print(f"Failed to write UART Modbus baudrate: {e}")
    def get_i2c_slave_address(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the I2C Slave Address from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: i2C slave address as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=I2C_SLAVE_ADDRESS)
        except Exception as e:
            print(f"Failed to read I2C slave address: {e}")
            return -1
    def set_i2c_slave_address(self, i2c_slave_address: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the I2C Slave Address to registers.

        :param i2c_slave_address: I2C slave address as a 16-bit unsigned integer value.
        :type i2c_slave_address: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=I2C_SLAVE_ADDRESS, value=i2c_slave_address)
        except Exception as e:
            print(f"Failed to write I2C slave address: {e}")
    def get_calibration_date(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Calibration Date from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Calibration date as a 32-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_long(registeraddress=CALIBRATION_DATE)
        except Exception as e:
            print(f"Failed to read calibration date: {e}")
            return -1
    def set_calibration_date(self, calibration_date: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Calibration Date to registers.

        :param calibration_date: Calibration date as a 32-bit unsigned integer value.
        :type calibration_date: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_long(registeraddress=CALIBRATION_DATE, value=calibration_date)
        except Exception as e:
            print(f"Failed to write calibration date: {e}")
    def get_calibration_response_factor(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Calibration Response Factor from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Calibration response factor as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=CALIBRATION_RESPONSE_FACTOR)
        except Exception as e:
            print(f"Failed to read calibration response factor: {e}")
            return -1.0
    def set_calibration_response_factor(self, calibration_response_factor: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Calibration Response Factor to registers.

        :param calibration_response_factor: Calibration response factor as a 32-bit float value.
        :type calibration_response_factor: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=CALIBRATION_RESPONSE_FACTOR, value=calibration_response_factor)
        except Exception as e:
            print(f"Failed to write calibration response factor: {e}")
    def get_zero_response(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Zero Response from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Zero response as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=ZERO_RESPONSE)
        except Exception as e:
            print(f"Failed to read zero response: {e}")
            return -1.0
    def set_zero_response(self, zero_response: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Zero Response to registers.

        :param zero_response: Zero response as a 32-bit float value.
        :type zero_response: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=ZERO_RESPONSE, value=zero_response)
        except Exception as e:
            print(f"Failed to write zero response: {e}")
    def get_span_1_concentration(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Span 1 Concentration from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Span 1 concentration as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=SPAN_1_CONCENTRATION)
        except Exception as e:
            print(f"Failed to read span 1 concentration: {e}")
            return -1.0
    def set_span_1_concentration(self, span_1_concentration: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Span 1 Concentration to registers.

        :param span_1_concentration: Span 1 concentration as a 32-bit float value.
        :type span_1_concentration: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=SPAN_1_CONCENTRATION, value=span_1_concentration)
        except Exception as e:
            print(f"Failed to write span 1 concentration: {e}")
    def get_span_1_response(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Span 1 Response from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Span 1 response as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=SPAN_1_RESPONSE)
        except Exception as e:
            print(f"Failed to read span 1 response: {e}")
            return -1.0
    def set_span_1_response(self, span_1_response: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Span 1 Response to registers.

        :param span_1_response: Span 1 response as a 32-bit float value.
        :type span_1_response: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=SPAN_1_RESPONSE, value=span_1_response)
        except Exception as e:
            print(f"Failed to write span 1 response: {e}")
    def get_span_2_concentration(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Span 2 Concentration from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Span 2 concentration as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=SPAN_2_CONCENTRATION)
        except Exception as e:
            print(f"Failed to read span 2 concentration: {e}")
            return -1.0
    def set_span_2_concentration(self, span_2_concentration: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Span 2 Concentration to registers.

        :param span_2_concentration: Span 2 concentration as a 32-bit float value.
        :type span_2_concentration: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=SPAN_2_CONCENTRATION, value=span_2_concentration)
        except Exception as e:
            print(f"Failed to write span 2 concentration: {e}")
    def get_span_2_response(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Span 2 Response from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Span 2 response as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=SPAN_2_RESPONSE)
        except Exception as e:
            print(f"Failed to read span 2 response: {e}")
            return -1.0
    def set_span_2_response(self, span_2_response: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Span 2 Response to registers.

        :param span_2_response: Span 2 response as a 32-bit float value.
        :type span_2_response: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=SPAN_2_RESPONSE, value=span_2_response)
        except Exception as e:
            print(f"Failed to write span 2 response: {e}")
    def get_span_3_concentration(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Span 3 Concentration from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Span 3 concentration as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=SPAN_3_CONCENTRATION)
        except Exception as e:
            print(f"Failed to read span 3 concentration: {e}")
            return -1.0
    def set_span_3_concentration(self, span_3_concentration: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Span 3 Concentration to registers.

        :param span_3_concentration: Span 3 concentration as a 32-bit float value.
        :type span_3_concentration: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=SPAN_3_CONCENTRATION, value=span_3_concentration)
        except Exception as e:
            print(f"Failed to write span 3 concentration: {e}")
    def get_span_3_response(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Span 3 Response from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Span 3 response as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=SPAN_3_RESPONSE)
        except Exception as e:
            print(f"Failed to read span 3 response: {e}")
            return -1.0
    def set_span_3_response(self, span_3_response: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Span 3 Response to registers.

        :param span_3_response: Span 3 response as a 32-bit float value.
        :type span_3_response: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=SPAN_3_RESPONSE, value=span_3_response)
        except Exception as e:
            print(f"Failed to write span 3 response: {e}")
    def get_span_4_concentration(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Span 4 Concentration from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Span 4 concentration as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=SPAN_4_CONCENTRATION)
        except Exception as e:
            print(f"Failed to read span 4 concentration: {e}")
            return -1.0
    def set_span_4_concentration(self, span_4_concentration: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Span 4 Concentration to registers.

        :param span_4_concentration: Span 4 concentration as a 32-bit float value.
        :type span_4_concentration: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=SPAN_4_CONCENTRATION, value=span_4_concentration)
        except Exception as e:
            print(f"Failed to write span 4 concentration: {e}")
    def get_span_4_response(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Span 4 Response from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Span 4 response as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=SPAN_4_RESPONSE)
        except Exception as e:
            print(f"Failed to read span 4 response: {e}")
            return -1.0
    def set_span_4_response(self, span_4_response: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Span 4 Response to registers.

        :param span_4_response: Span 4 response as a 32-bit float value.
        :type span_4_response: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=SPAN_4_RESPONSE, value=span_4_response)
        except Exception as e:
            print(f"Failed to write span 4 response: {e}")
    def get_sensor_mv(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Sensor MV Value from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Sensor MV value as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return round(self._modbus.read_float(registeraddress=SENSOR_MV_VALUE), 4)
        except Exception as e:
            print(f"Failed to read sensor MV value: {e}")
            return -1.0
    def get_sensor_ppm(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Sensor PPM from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Sensor PPM as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return round(self._modbus.read_float(registeraddress=SENSOR_PPM_VALUE), 4)
        except Exception as e:
            print(f"Failed to read sensor PPM: {e}")
            return -1.0
    def get_temperature(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Temperature from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Temperature as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=TEMPERATURE)
        except Exception as e:
            print(f"Failed to read temperature: {e}")
            return -1.0
    def get_humidity(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Humidity from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Humidity as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=HUMIDITY)
        except Exception as e:
            print(f"Failed to read humidity: {e}")
            return -1.0
    def get_pressure(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Pressure from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Pressure as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=PRESSURE)
        except Exception as e:
            print(f"Failed to read pressure: {e}")
            return -1.0
    def get_4ma_current_loop(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the 4mA Current Loop from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: The 4mA current loop as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=CURRENT_LOOP_4MA)
        except Exception as e:
            print(f"Failed to read 4mA current loop: {e}")
            return -1.0
    def get_20ma_current_loop(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the 20mA Current Loop from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: The 20mA current loop as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=CURRENT_LOOP_20MA)
        except Exception as e:
            print(f"Failed to read 20mA current loop: {e}")
            return -1.0
    def get_420ma_current_loop_sensitivity(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the 4-20mA Current Loop Sensitivity from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: The 4-20mA current loop sensitivity as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=CURRENT_LOOP_4_20MA_SENSITIVITY)
        except Exception as e:
            print(f"Failed to read 4-20mA current loop sensitivity: {e}")
            return -1.0
    def get_dac_value(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the DAC Value from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: DAC value as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=DAC_VALUE)
        except Exception as e:
            print(f"Failed to read DAC value: {e}")
            return -1.0
    def get_dac_data(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the DAC Data from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: DAC data as a 32-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=DAC_DATA)
        except Exception as e:
            print(f"Failed to read DAC data: {e}")
            return -1
    def get_working_status(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Working Status from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Working state as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=WORKING_STATUS)
        except Exception as e:
            print(f"Failed to read working status: {e}")
            return -1
    def get_sensor_error_status(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Sensor Error Status from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Sensor error status as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=SENSOR_ERROR_STATUS)
        except Exception as e:
            print(f"Failed to read sensor error status: {e}")
            return -1
    def get_pid_supply(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the PID Supply from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: PID supply as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=PID_SUPPLY)
        except Exception as e:
            print(f"Failed to read PID supply: {e}")
            return -1
    def set_pid_supply(self, pid_supply: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the PID Supply to registers.

        :param pid_supply: PID supply as a 16-bit unsigned integer value.
        :type pid_supply: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=PID_SUPPLY, value=pid_supply)
        except Exception as e:
            print(f"Failed to write PID supply: {e}")
    def get_alarm_latch_clear(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Alarm Latch Clear from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Alarm latch clear as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=ALARM_LATCH_CLEAR)
        except Exception as e:
            print(f"Failed to read PID supply: {e}")
            return -1    
    def set_alarm_latch_clear(self, alarm_latch_clear: int = 1, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Alarm Latch Clear to registers.

        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=ALARM_LATCH_CLEAR, value=alarm_latch_clear)
        except Exception as e:
            print(f"Failed to write alarm latch clear: {e}")
    def get_operation_mode(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Operation Mode from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Operation mode as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=OPERATION_MODE)
        except Exception as e:
            print(f"Failed to read operation mode: {e}")
            return -1
    def set_operation_mode(self, operation_mode: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Operation Mode to registers.

        :param operation_mode: Operation mode as a 16-bit unsigned integer value.
        :type operation_mode: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=OPERATION_MODE, value=operation_mode)
        except Exception as e:
            print(f"Failed to write operation mode: {e}")
    def get_relative_zero_value(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> float:
        """
        Reads the Relative Zero Value from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Relative zero value as a 32-bit float value.
        :rtype: float
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_float(registeraddress=NEW_RELATIVE_ZERO)
        except Exception as e:
            print(f"Failed to read relative zero value: {e}")
            return -1.0
    def set_relative_zero_value(self, relative_zero_value: float, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Relative Zero Value to registers.

        :param relative_zero_value: Relative zero value as a 32-bit float value.
        :type relative_zero_value: float
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_float(registeraddress=NEW_RELATIVE_ZERO, value=relative_zero_value)
        except Exception as e:
            print(f"Failed to write relative zero value: {e}")
    def get_relative_zero_state(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> int:
        """
        Reads the Relative Zero State from registers.

        :param slave_address: Slave address to query from.
        :type slave_address: int
        :returns: Relative zero state as a 16-bit unsigned integer value.
        :rtype: int
        """
        self._set_address(slave_address)
        try:
            return self._modbus.read_register(registeraddress=USE_NEW_RELATIVE_ZERO)
        except Exception as e:
            print(f"Failed to read relative zero state: {e}")
            return -1
    def set_relative_zero_state(self, relative_zero_state: int, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Relative Zero State to registers.

        :param relative_zero_state: Relative zero state as a 16-bit unsigned integer value.
        :type relative_zero_state: int
        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=USE_NEW_RELATIVE_ZERO, value=relative_zero_state)
        except Exception as e:
            print(f"Failed to write relative zero state: {e}")
    def set_software_reset(self, slave_address: int = DEFAULT_MODBUS_ADDRESS) -> None:
        """
        Writes the Software Reset to registers.

        :param slave_address: Slave address to write to.
        :type slave_address: int
        """
        self._set_address(slave_address)
        try:
            self._modbus.write_register(registeraddress=SOFTWARE_RESET, value=0)
            self._modbus.write_register(registeraddress=SOFTWARE_RESET, value=1)
        except Exception as e:
            print(f"Failed to write software reset: {e}")

class I2C:
    """Base functionality for I2C Protocol

    **Quickstart: Importing and using Ion_Connect I2C Protocol**

        Here is an example of using the :class:`I2C`.
        First you will need to import the libraries to use the sensor

        .. code-block:: python

            import ION_Connect

        Once this is done you can define your `Ion_Connect.I2C` object and define the object.

        .. code-block:: python

            i2c_address = 0x15
            i2c_bus = 1
            ion_i2c = Ion_Connect.I2C(i2c_address, i2c_bus)

        Now you have access to the registers.
        The following example will read the sensor mV and return a float to :variable:'sensor_mv'.

        .. code-block:: python

            sensor_mv = ion_i2c.get_sensor_mv()

        The following example will set the response factor to value 2 and return a float to :variable:'response_factor'.
        In order to set the response factor value, the operation mode requires to be set to 1 in order to access configure mode.

        .. code-block:: python

            ion_i2c.set_operation_mode(1)
            ion_i2c.set_response_factor(2)
            ion_i2c.set_operation_mode(0)

            response_factor = ion_i2c.get_response_factor()

    """
    def __init__(self, i2c_address: int = DEFAULT_I2C_ADDRESS, i2c_bus: int = None) -> None:
        """
        Initialisation of the I2C communication protocol.

        :param i2c_address: I2C address to query from.
        :type i2c_address: int
        :param i2c_bus: I2C bus to query from.
        :type i2c_bus: int
        """
        if i2c_bus is None:
            raise ConnectionError("No I2C Bus has been defined!")
        self._i2c_address = i2c_address
        self._i2c_bus = i2c_bus
        self._bus = None
    @property
    def i2c_address(self) -> int:
        """
        Gets the I2C address.

        :returns: I2C address value.
        :rtype: int
        """
        return self._i2c_address
    @property
    def i2c_bus(self) -> int:
        """
        Gets the I2C bus.        
        
        :returns: I2C bus value.
        :rtype: int
        """
        return self._i2c_bus   
    def open(self) -> None:
        """Opens the I2C bus."""
        if self._bus is None:
            try:
                self._bus = SMBus.SMBus(self._i2c_bus)
            except Exception as e:
                raise ConnectionError(f"Failed to open I2C bus: {e}")
    def close(self) -> None:
        """Closes the I2C bus."""
        if self._bus is not None:
            try:
                self._bus.close()
            except Exception as e:
                print(f"Failed to close I2C bus: {e}")
            finally:
                self._bus = None
    def __del__(self) -> None:
        """Ensures the I2C bus is closed when the object is deleted."""
        self.close()
    def flush(self) -> None:
        """Flushes the I2C buffer."""
        self.close()
        self.open()

    def _write_8_bit(self, i2c_register: int, data: int) -> None:
        """
        Writes the data as an 8-bit unsigned integer.

        :param i2c_register: I2C register to write to.
        :type i2c_register: int
        :param data: Data to write to the register.
        :type data: int
        """
        payload = [i2c_register, data]
        payload.append(self._calculate_crc8(payload, 3))
        try:
            self._bus.write_i2c_block_data(i2c_addr=self._i2c_address, register=i2c_register, data=payload[1:4])
        except Exception as e:
            print(f"Failed to write 8-bit data: {e}")
    def _write_32_uint(self, i2c_register: int, data: int) -> None:
        """
        Writes the data as a 32-bit unsigned integer.

        :param i2c_register: I2C register to write to.
        :type i2c_register: int
        :param data: Data to write to the register.
        :type data: int
        """
        payload = list(self._uint32_to_bytes(data))
        payload.insert(0, i2c_register)
        payload.append(self._calculate_crc8(payload, 6))
        try:
            self._bus.write_i2c_block_data(i2c_addr=self._i2c_address, register=i2c_register, data=payload[1:6])
        except Exception as e:
            print(f"Failed to write 32-bit unsigned integer: {e}")
    def _write_32_float(self, i2c_register: int, data: float) -> None:
        """
        Writes the data as a 32-bit float.

        :param i2c_register: I2C register to write to.
        :type i2c_register: int
        :param data: Data to write to the register.
        :type data: float
        """
        payload = list(self._float32_to_bytes(data))
        payload.insert(0, i2c_register)
        payload.append(self._calculate_crc8(payload, 6))
        try:
            self._bus.write_i2c_block_data(i2c_addr=self._i2c_address, register=i2c_register, data=payload[1:6])
        except Exception as e:
            print(f"Failed to write 32-bit float: {e}")
    def _read_8_bit(self, i2c_register: int) -> int:
        """
        Reads the data as an 8-bit unsigned integer.

        :param i2c_register: I2C register to query from.
        :type i2c_register: type
        :returns: Value from I2C register as an 8-bit unsigned integer.
        :rtype: int
        """
        try:
            buffer = self._bus.read_i2c_block_data(i2c_addr=self._i2c_address, register=i2c_register, length=3)
            if self._calculate_crc8(buffer, 3) == buffer[2]:
                return buffer[1]
            else:
                raise ValueError("Checksum Error")
        except Exception as e:
            print(f"Failed to read 8-bit data: {e}")
            return -1
    def _read_16_bit(self, i2c_register: int) -> int:
        """
        Reads the data as a 16-bit unsigned integer.

        :param i2c_register: I2C register to query from.
        :type i2c_register: type
        :returns: Value from I2C register as a 16-bit unsigned integer.
        :rtype: int
        """
        try:
            buffer = self._bus.read_i2c_block_data(i2c_addr=self._i2c_address, register=i2c_register, length=4)
            if self._calculate_crc8(buffer, 4) == buffer[3]:
                return self._bytes_to_uint16(buffer[1:3])
            else:
                raise ValueError("Checksum Error")
        except Exception as e:
            print(f"Failed to read 16-bit data: {e}")
            return -1
    def _read_32_uint(self, i2c_register: int) -> int:
        """
        Reads the data as a 32-bit unsigned integer.

        :param i2c_register: I2C register to query from.
        :type i2c_register: type
        :returns: Value from I2C register as a 32-bit unsigned integer.
        :rtype: int
        """
        try:
            buffer = self._bus.read_i2c_block_data(i2c_addr=self._i2c_address, register=i2c_register, length=6)
            if self._calculate_crc8(buffer, 6) == buffer[5]:
                return self._bytes_to_uint32(bytes(buffer[1:5]))
            else:
                raise ValueError("Checksum Error")
        except Exception as e:
            print(f"Failed to read 32-bit unsigned integer: {e}")
            return -1
    def _read_32_float(self, i2c_register: int) -> float:
        """
        Reads the data as a 32-bit float.

        :param i2c_register: I2C register to query from.
        :type i2c_register: type
        :returns: Value from I2C register as a 32-bit float.
        :rtype: float
        """
        try:
            buffer = self._bus.read_i2c_block_data(i2c_addr=self._i2c_address, register=i2c_register, length=6)
            if self._calculate_crc8(buffer, 6) == buffer[5]:
                return self._bytes_to_float32(bytes(buffer[1:5]))
            else:
                raise ValueError("Checksum Error")
        except Exception as e:
            print(f"Failed to read 32-bit float: {e}")
            return -1.0
    @staticmethod
    def _calculate_crc8(data: list, size: int) -> int:
        """
        Calculates CRC-8/NRSC-5.

        :param data: Data received from buffer.
        :type data: list
        :param size: Byte size of buffer.
        :type: size: int
        :returns: CRC value calculated from buffer using CRC-8/NRSC-5.
        :rtype: int
        """
        if size == 3:
            fdata = data[0:2]
        elif size == 4:
            fdata = data[0:3]
        elif size == 6:
            fdata = data[0:5]
        else:
            raise ValueError("Unsupported size for CRC calculation")

        crc = 0xff
        for byte in fdata:
            crc ^= byte
            for _ in range(8):
                if (crc & 0x80) != 0:
                    crc = (crc << 1) ^ 0x31
                else:
                    crc <<= 1
                crc &= 0xff  # Ensure crc remains an 8-bit value
        return crc
    @staticmethod
    def _uint16_to_bytes(uint16_value: int) -> bytes:
        """
        Converts a 16-bit unsigned integer to a sequence of bytes.

        :param uint16_value: A 16-bit unsigned integer value.
        :type uint16_value: int
        :returns: A bytes object representing the 16-bit unsigned integer.
        :rtype: bytes
        :raises ValueError: If the input is not a valid 16-bit unsigned integer (0 <= value <= 65535).
        """
        if not (0 <= uint16_value <= 0xFFFF):
            raise ValueError("Input must be a uint16 (0 <= value <= 65535)")

        return uint16_value.to_bytes(length=2, byteorder='big')
    @staticmethod
    def _uint32_to_bytes(uint32_value: int) -> bytes:
        """
        Converts a 32-bit unsigned integer to a sequence of bytes.

        :param uint32_value: A 32-bit unsigned integer value.
        :type uint32_value: int
        :returns: A bytes object representing the 32-bit unsigned integer.
        :rtype: bytes
        :raises ValueError: If the input is not a valid 32-bit unsigned integer (0 <= value <= 4294967295).
        """
        if not (0 <= uint32_value <= 0xFFFFFFFF):
            raise ValueError("Input must be a uint32 (0 <= value <= 4294967295)")

        return uint32_value.to_bytes(4, byteorder='big', signed=False)
    @staticmethod
    def _float32_to_bytes(float32_value: float) -> bytes:
        """
        Converts a 32-bit floating-point number to a sequence of bytes.

        :param float32_value: A 32-bit floating-point value.
        :type float32_value: float
        :returns: A bytes object representing the 32-bit floating-point number.
        :rtype: bytes
        """
        if float32_value != float32_value:  # NaN check
            return b'\x7f\xc0\x00\x00'
        elif float32_value == float('inf'):
            return b'\x7f\x80\x00\x00'
        elif float32_value == float('-inf'):
            return b'\xff\x80\x00\x00'

        sign = 0 if float32_value >= 0 else 1
        float32_value = abs(float32_value)

        if float32_value == 0.0:
            exponent = 0
            mantissa = 0
        else:
            exponent = 127
            while float32_value >= 2.0:
                float32_value /= 2.0
                exponent += 1
            while float32_value < 1.0:
                float32_value *= 2.0
                exponent -= 1

            float32_value -= 1.0
            mantissa = int(float32_value * (2**23))

        int_value = (sign << 31) | (exponent << 23) | (mantissa & 0x7FFFFF)
        return int_value.to_bytes(4, byteorder='big')
    @staticmethod
    def _bytes_to_uint16(byte_data: bytes) -> int:
        """
        Converts a sequence of bytes to a 16-bit unsigned integer.

        :param byte_data: A bytes object containing exactly 2 bytes.
        :type byte_data: bytes
        :returns: A 16-bit unsigned integer derived from the byte data.
        :rtype: int
        :raises ValueError: If the input does not contain exactly 2 bytes.
        """
        if len(byte_data) != 2:
            raise ValueError("Input must be exactly 2 bytes")

        return int.from_bytes(byte_data, byteorder='big', signed=False)
    @staticmethod
    def _bytes_to_uint32(byte_data: bytes) -> int:
        """
        Converts a sequence of bytes to a 32-bit unsigned integer.

        :param byte_data: A bytes object containing exactly 4 bytes.
        :type byte_data: bytes
        :returns: A 32-bit unsigned integer derived from the byte data.
        :rtype: int
        :raises ValueError: If the input does not contain exactly 4 bytes.
        """
        if len(byte_data) != 4:
            raise ValueError("Input must be exactly 4 bytes")

        return int.from_bytes(byte_data, byteorder='big', signed=False)
    @staticmethod
    def _bytes_to_float32(byte_data: bytes) -> float:
        """
        Converts a sequence of bytes to a 32-bit floating-point number.

        :param byte_data: A bytes object containing exactly 4 bytes.
        :type byte_data: bytes
        :returns: A 32-bit floating-point number derived from the byte data.
        :rtype: float
        :raises ValueError: If the input does not contain exactly 4 bytes.
        """
        if len(byte_data) != 4:
            raise ValueError("Input must be exactly 4 bytes")

        int_value = int.from_bytes(byte_data, byteorder='big', signed=False)
        sign = (int_value >> 31) & 0x01
        exponent = (int_value >> 23) & 0xFF
        mantissa = int_value & 0x7FFFFF

        if exponent == 0:
            float32_value = (-1)**sign * 2**(-126) * (mantissa / 2**23)
        elif exponent == 0xFF:
            if mantissa == 0:
                float32_value = float('inf') if sign == 0 else float('-inf')
            else:
                float32_value = float('nan')
        else:
            float32_value = (-1)**sign * 2**(exponent - 127) * (1 + mantissa / 2**23)

        return float32_value

    def get_board_serial_number(self) -> int:
        """
        Reads the Board Serial Number via I2C.

        :returns: Board serial number as a 32-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_32_uint(i2c_register=BOARD_SERIAL_NUMBER)
        except Exception as e:
            print(f"Failed to read board serial number: {e}")
            return -1
    def get_board_batch_number(self) -> int:
        """
        Reads the Board Batch Number via I2C.

        :returns: Board batch number as a 32-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_32_uint(i2c_register=BOARD_BATCH_NUMBER)
        except Exception as e:
            print(f"Failed to read board batch number: {e}")
            return -1
    def get_production_date(self) -> int:
        """
        Reads the Production Date via I2C.

        :returns: Production date as a 32-bit unsigned integer value in format YYYYMMDD.
        :rtype: int
        """
        try:
            return self._read_32_uint(i2c_register=PRODUCTION_DATE)
        except Exception as e:
            print(f"Failed to read production date: {e}")
            return -1
    def get_pcb_hardware(self) -> int:
        """
        Reads the PCB Hardware version via I2C.

        :returns: PCB hardware version as a 16-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_16_bit(i2c_register=PCB_HARDWARE_VERSION)
        except Exception as e:
            print(f"Failed to read PCB hardware version: {e}")
            return -1
    def get_firmware_version(self) -> int:
        """
        Reads the Firmware Version via I2C.

        :returns: Firmware version as a 16-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_16_bit(i2c_register=FIRMWARE_VERSION)
        except Exception as e:
            print(f"Failed to read firmware version: {e}")
            return -1
    def get_sensor_type(self) -> int:
        """
        Reads the Sensor Type via I2C.

        :returns: Sensor type as a 16-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=SENSOR_TYPE)
        except Exception as e:
            print(f"Failed to read sensor type: {e}")
            return -1
    def set_sensor_type(self, sensor_type: int) -> None:
        """
        Writes the Sensor Type via I2C.

        :param sensor_type: Sensor type as a 16-bit unsigned integer value.
        :type sensor_type: int
        """
        try:
            self._write_8_bit(i2c_register=SENSOR_TYPE, data=sensor_type)
        except Exception as e:
            print(f"Failed to write sensor type: {e}")
    def get_response_factor(self) -> float:
        """
        Reads the Response Factor via I2C.

        :returns: Response factor as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=RESPONSE_FACTOR_VALUE)
        except Exception as e:
            print(f"Failed to read response factor: {e}")
            return -1.0
    def set_response_factor(self, response_factor: float) -> None:
        """
        Writes the Response Factor via I2C.

        :param response_factor: Response factor as a 32-bit float value.
        :type response_factor: float
        """
        try:
            self._write_32_float(i2c_register=RESPONSE_FACTOR_VALUE, data=response_factor)
        except Exception as e:
            print(f"Failed to write response factor: {e}")
    def get_sensor_maximum_value(self) -> float:
        """
        Reads the Sensor Maximum Value via I2C.

        :returns: Sensor maximum value as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=SENSOR_MAXIMUM_VALUE)
        except Exception as e:
            print(f"Failed to read sensor maximum value: {e}")
            return -1.0
    def set_sensor_maximum_value(self, sensor_maximum: float) -> None:
        """
        Writes the Sensor Maximum Value via I2C.

        :param sensor_maximum: Sensor maximum value as a 32-bit float value.
        :type sensor_maximum: float
        """
        try:
            self._write_32_float(i2c_register=SENSOR_MAXIMUM_VALUE, data=sensor_maximum)
        except Exception as e:
            print(f"Failed to write sensor maximum value: {e}")
    def get_sensor_maximum_state(self) -> int:
        """
        Reads the Sensor Maximum State via I2C.

        :returns: Sensor maximum state as a 16-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=USE_SENSOR_MAXIMUM_VALUE)
        except Exception as e:
            print(f"Failed to read sensor maximum state: {e}")
            return -1
    def set_sensor_maximum_state(self, sensor_maximum_state: int) -> None:
        """
        Writes the Sensor Maximum State via I2C.

        :param sensor_maximum_state: Sensor maximum state as a 16-bit unsigned integer value.
        :type sensor_maxmimum_state: int
        """
        try:
            self._write_8_bit(i2c_register=USE_SENSOR_MAXIMUM_VALUE, data=sensor_maximum_state)
        except Exception as e:
            print(f"Failed to write sensor maximum state: {e}")
    def get_alarm_threshold_value(self) -> float:
        """
        Reads the Alarm Threshold Value via I2C.

        :returns: Alarm threshold value as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=ALARM_THRESHOLD_VALUE)
        except Exception as e:
            print(f"Failed to read alarm threshold value: {e}")
            return -1.0
    def set_alarm_threshold_value(self, alarm_threshold_value: float) -> None:
        """
        Writes the Alarm Threshold Value via I2C.
        
        :param alarm_threshold_value: Alarm threshold value as a 32-bit float value.
        :type alarm_threshold_value: float
        """
        try:
            self._write_32_float(i2c_register=ALARM_THRESHOLD_VALUE, data=alarm_threshold_value)
        except Exception as e:
            print(f"Failed to write alarm threshold value: {e}")
    def get_alarm_latch_state(self) -> int:
        """
        Reads the Alarm Latch state via I2C.

        :returns: Alarm latch as a 16-bit unsigned integer value.
        :rtype: ints
        """
        try:
            return self._read_8_bit(i2c_register=ALARM_LATCH)
        except Exception as e:
            print(f"Failed to read alarm latch state: {e}")
            return -1
    def set_alarm_latch_state(self, alarm_latch_state: int) -> None:
        """
        Writes the Alarm Latch State via I2C.

        :param alarm_latch_state: Alarm latch state as a 16-bit unsigned integer value.
        :type alarm_latch_state: int
        """
        try:
            self._write_8_bit(i2c_register=ALARM_LATCH, data=alarm_latch_state)
        except Exception as e:
            print(f"Failed to write alarm latch state: {e}")
    def get_led_control(self) -> int:
        """
        Reads the LED Control via I2C.

        :returns: LED control as a 16-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=LED_CONTROL)
        except Exception as e:
            print(f"Failed to read LED control: {e}")
            return -1
    def set_led_control(self, led_control: int) -> None:
        """
        Writes the LED Control Status via I2C.

        :param led_control: LED control state as a 16-bit unsigned integer value.
        :type led_control: int
        """
        try:
            self._write_8_bit(i2c_register=LED_CONTROL, data=led_control)
        except Exception as e:
            print(f"Failed to write LED control: {e}")
    def get_modbus_slave_address(self) -> int:
        """
        Reads the Modbus Slave Address via I2C.

        :returns: Modbus slave address as a 16-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=MODBUS_SLAVE_ADDRESS)
        except Exception as e:
            print(f"Failed to read Modbus slave address: {e}")
            return -1
    def set_modbus_slave_address(self, modbus_slave_address: int) -> None:
        """
        Writes the Modbus Slave Address via I2C.

        :param modbus_slave_address: Modbus slave address as a 16-bit unsigned integer value.
        :type modbus_slave_address: int
        """
        try:
            self._write_8_bit(i2c_register=MODBUS_SLAVE_ADDRESS, data=modbus_slave_address)
        except Exception as e:
            print(f"Failed to write Modbus slave address: {e}")
    def get_uart_modbus_baudrate(self) -> int:
        """
        Reads the UART Modbus Baudrate via I2C.

        :returns: UART Modbus baudrate as a 16-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=UART_MODBUS_BAUD_RATE)
        except Exception as e:
            print(f"Failed to read UART Modbus baudrate: {e}")
            return -1
    def set_uart_modbus_baudrate(self, uart_modbus_baudrate: int) -> None:
        """
        Writes the UART Modbus Baudrate via I2C.

        :param uart_modbus_baudrate: UART Modbus baudrate as a 16-bit unsigned integer value.
        :type uart_modbus_baudrate: int
        """
        try:
            self._write_8_bit(i2c_register=UART_MODBUS_BAUD_RATE, data=uart_modbus_baudrate)
        except Exception as e:
            print(f"Failed to write UART Modbus baudrate: {e}")
    def get_i2c_slave_address(self) -> int:
        """
        Reads the I2C Slave Address via I2C.

        :returns: i2C slave address as a 16-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=I2C_SLAVE_ADDRESS)
        except Exception as e:
            print(f"Failed to read I2C slave address: {e}")
            return -1
    def set_i2c_slave_address(self, i2c_slave_address: int) -> None:
        """
        Writes the I2C Slave Address via I2C.

        :param i2c_slave_address: I2C slave address as a 16-bit unsigned integer value.
        :type i2c_slave_address: int
        """
        try:
            self._write_8_bit(i2c_register=I2C_SLAVE_ADDRESS, data=i2c_slave_address)
        except Exception as e:
            print(f"Failed to write I2C slave address: {e}")
    def get_calibration_date(self) -> int:
        """
        Reads the Calibration Date via I2C.

        :returns: Calibration date as a 32-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_32_uint(i2c_register=CALIBRATION_DATE)
        except Exception as e:
            print(f"Failed to read calibration date: {e}")
            return -1
    def set_calibration_date(self, calibration_date: int) -> None:
        """
        Writes the Calibration Date via I2C.

        :param calibration_date: Calibration date as a 32-bit unsigned integer value.
        :type calibration_date: int
        """
        try:
            self._write_32_uint(i2c_register=CALIBRATION_DATE, data=calibration_date)
        except Exception as e:
            print(f"Failed to write calibration date: {e}")
    def get_calibration_response_factor(self) -> float:
        """
        Reads the Calibration Response Factor via I2C.

        :returns: Calibration response factor as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=CALIBRATION_RESPONSE_FACTOR)
        except Exception as e:
            print(f"Failed to read calibration response factor: {e}")
            return -1.0
    def set_calibration_response_factor(self, calibration_response_factor: float) -> None:
        """
        Writes the Calibration Response Factor via I2C.

        :param calibration_response_factor: Calibration response factor as a 32-bit float value.
        :type calibration_response_factor: float
        """
        try:
            self._write_32_float(i2c_register=CALIBRATION_RESPONSE_FACTOR, data=calibration_response_factor)
        except Exception as e:
            print(f"Failed to write calibration response factor: {e}")
    def get_zero_response(self) -> float:
        """
        Reads the Zero Response via I2C.

        :returns: Zero response as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=ZERO_RESPONSE)
        except Exception as e:
            print(f"Failed to read zero response: {e}")
            return -1.0
    def set_zero_response(self, zero_response: float) -> None:
        """
        Writes the Zero Response via I2C.

        :param zero_response: Zero response as a 32-bit float value.
        :type zero_response: float
        """
        try:
            self._write_32_float(i2c_register=ZERO_RESPONSE, data=zero_response)
        except Exception as e:
            print(f"Failed to write zero response: {e}")
    def get_span_1_concentration(self) -> float:
        """
        Reads the Span 1 Concentration via I2C.

        :returns: Span 1 concentration as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=SPAN_1_CONCENTRATION)
        except Exception as e:
            print(f"Failed to read span 1 concentration: {e}")
            return -1.0
    def set_span_1_concentration(self, span_1_concentration: float) -> None:
        """
        Writes the Span 1 Concentration via I2C.

        :param span_1_concentration: Span 1 concentration as a 32-bit float value.
        :type span_1_concentration: float
        """
        try:
            self._write_32_float(i2c_register=SPAN_1_CONCENTRATION, data=span_1_concentration)
        except Exception as e:
            print(f"Failed to write span 1 concentration: {e}")
    def get_span_1_response(self) -> float:
        """
        Reads the Span 1 Response via I2C.

        :returns: Span 1 response as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=SPAN_1_RESPONSE)
        except Exception as e:
            print(f"Failed to read span 1 response: {e}")
            return -1.0
    def set_span_1_response(self, span_1_response: float) -> None:
        """
        Writes the Span 1 Response via I2C.

        :param span_1_response: Span 1 response as a 32-bit float value.
        :type span_1_response: float
        """
        try:
            self._write_32_float(i2c_register=SPAN_1_RESPONSE, data=span_1_response)
        except Exception as e:
            print(f"Failed to write span 1 response: {e}")
    def get_span_2_concentration(self) -> float:
        """
        Reads the Span 2 Concentration via I2C.

        :returns: Span 2 concentration as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=SPAN_2_CONCENTRATION)
        except Exception as e:
            print(f"Failed to read span 2 concentration: {e}")
            return -1.0
    def set_span_2_concentration(self, span_2_concentration: float) -> None:
        """
        Writes the Span 2 Concentration via I2C.

        :param span_2_concentration: Span 2 concentration as a 32-bit float value.
        :type span_2_concentration: float
        """
        try:
            self._write_32_float(i2c_register=SPAN_2_CONCENTRATION, data=span_2_concentration)
        except Exception as e:
            print(f"Failed to write span 2 concentration: {e}")
    def get_span_2_response(self) -> float:
        """
        Reads the Span 2 Response via I2C.

        :returns: Span 2 response as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=SPAN_2_RESPONSE)
        except Exception as e:
            print(f"Failed to read span 2 response: {e}")
            return -1.0
    def set_span_2_response(self, span_2_response: float) -> None:
        """
        Writes the Span 2 Response via I2C.

        :param span_2_response: Span 2 response as a 32-bit float value.
        :type span_2_response: float
        """
        try:
            self._write_32_float(i2c_register=SPAN_2_RESPONSE, data=span_2_response)
        except Exception as e:
            print(f"Failed to write span 2 response: {e}")
    def get_span_3_concentration(self) -> float:
        """
        Reads the Span 3 Concentration via I2C.

        :returns: Span 3 concentration as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=SPAN_3_CONCENTRATION)
        except Exception as e:
            print(f"Failed to read span 3 concentration: {e}")
            return -1.0    
    def set_span_3_concentration(self, span_3_concentration: float) -> None:
        """
        Writes the Span 3 Concentration via I2C.

        :param span_3_concentration: Span 3 concentration as a 32-bit float value.
        :type span_3_concentration: float
        """
        try:
            self._write_32_float(i2c_register=SPAN_3_CONCENTRATION, data=span_3_concentration)
        except Exception as e:
            print(f"Failed to write Span 3 concentration: {e}")
    def get_span_3_response(self) -> float:
        """
        Reads the Span 3 Response via I2C.

        :returns: Span 3 response as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=SPAN_3_RESPONSE)
        except Exception as e:
            print(f"Failed to read Span 3 response: {e}")
            return -1.0
    def set_span_3_response(self, span_3_response: float) -> None:
        """
        Writes the Span 3 Response via I2C.

        :param span_3_response: Span 3 response as a 32-bit float value.
        :type span_3_response: float
        """
        try:
            self._write_32_float(i2c_register=SPAN_3_RESPONSE, data=span_3_response)
        except Exception as e:
            print(f"Failed to write Span 3 response: {e}")
    def get_span_4_concentration(self) -> float:
        """
        Reads the Span 4 Concentration via I2C.

        :returns: Span 4 concentration as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=SPAN_4_CONCENTRATION)
        except Exception as e:
            print(f"Failed to read Span 4 concentration: {e}")
            return -1.0
    def set_span_4_concentration(self, span_4_concentration: float) -> None:
        """
        Writes the Span 4 Concentration via I2C.

        :param span_4_concentration: Span 4 concentration as a 32-bit float value.
        :type span_4_concentration: float
        """
        try:
            self._write_32_float(i2c_register=SPAN_4_CONCENTRATION, data=span_4_concentration)
        except Exception as e:
            print(f"Failed to write Span 4 concentration: {e}")
    def get_span_4_response(self) -> float:
        """
        Reads the Span 4 Response via I2C.

        :returns: Span 4 response as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=SPAN_4_RESPONSE)
        except Exception as e:
            print(f"Failed to read Span 4 response: {e}")
            return -1.0
    def set_span_4_response(self, span_4_response: float) -> None:
        """
        Writes the Span 4 Response via I2C.

        :param span_4_response: Span 4 response as a 32-bit float value.
        :type span_4_response: float
        """
        try:
            self._write_32_float(i2c_register=SPAN_4_RESPONSE, data=span_4_response)
        except Exception as e:
            print(f"Failed to write Span 4 response: {e}")
    def get_sensor_mv(self) -> float:
        """
        Reads the Sensor MV Value via I2C.

        :returns: Sensor MV Value as a 32-bit float value.
        :rtype: float
        """
        try:
            return round(self._read_32_float(i2c_register=SENSOR_MV_VALUE), 4)
        except Exception as e:
            print(f"Failed to read sensor MV value: {e}")
            return -1.0
    def get_sensor_ppm(self) -> float:
        """
        Reads the Sensor PPM via I2C.

        :returns: Sensor PPM as a 32-bit float value.
        :rtype: float
        """
        try:
            return round(self._read_32_float(i2c_register=SENSOR_PPM_VALUE), 4)
        except Exception as e:
            print(f"Failed to read sensor PPM: {e}")
            return -1.0
    def get_temperature(self) -> float:
        """
        Reads the Temperature via I2C.

        :returns: Temperature as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=TEMPERATURE)
        except Exception as e:
            print(f"Failed to read temperature: {e}")
            return -1.0
    def get_humidity(self) -> float:
        """
        Reads the Humidity via I2C.

        :returns: Humidity as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=HUMIDITY)
        except Exception as e:
            print(f"Failed to read humidity: {e}")
            return -1.0
    def get_pressure(self) -> float:
        """
        Reads the Pressure via I2C.

        :returns: Pressure as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=PRESSURE)
        except Exception as e:
            print(f"Failed to read pressure: {e}")
            return -1.0
    def get_4ma_current_loop(self) -> float:
        """
        Reads the 4mA Current Loop via I2C.

        :returns: 4mA current loop as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=CURRENT_LOOP_4MA)
        except Exception as e:
            print(f"Failed to read 4mA current loop: {e}")
            return -1.0
    def get_20ma_current_loop(self) -> float:
        """
        Reads the 20mA Current Loop via I2C.

        :returns: 20mA current loop as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=CURRENT_LOOP_20MA)
        except Exception as e:
            print(f"Failed to read 20mA current loop: {e}")
            return -1.0
    def get_420ma_current_loop_sensitivity(self) -> float:
        """
        Reads the 4-20mA Current Loop Sensitivity via I2C.

        :returns: 4-20mA current loop sensitivity as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=CURRENT_LOOP_4_20MA_SENSITIVITY)
        except Exception as e:
            print(f"Failed to read 4-20mA current loop sensitivity: {e}")
            return -1.0
    def get_dac_value(self) -> float:
        """
        Reads the DAC Value via I2C.

        :returns: DAC value as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=DAC_VALUE)
        except Exception as e:
            print(f"Failed to read DAC value: {e}")
            return -1.0
    def get_dac_data(self) -> int:
        """
        Reads the DAC Data via I2C.

        :returns: DAC data as a 16-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_16_bit(i2c_register=DAC_DATA)
        except Exception as e:
            print(f"Failed to read DAC data: {e}")
            return -1
    def get_working_status(self) -> int:
        """
        Reads the Working Status via I2C.

        :returns: Working status as an 8-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=WORKING_STATUS)
        except Exception as e:
            print(f"Failed to read working status: {e}")
            return -1
    def get_sensor_error_status(self) -> int:
        """
        Reads the Sensor Error Status via I2C.

        :returns: Sensor error status as an 8-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=SENSOR_ERROR_STATUS)
        except Exception as e:
            print(f"Failed to read sensor error status: {e}")
            return -1
    def get_pid_supply(self) -> int:
        """
        Reads the PID Supply via I2C.

        :returns: PID supply as an 8-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=PID_SUPPLY)
        except Exception as e:
            print(f"Failed to read PID supply: {e}")
            return -1
    def set_pid_supply(self, pid_supply: int) -> None:
        """
        Writes the PID Supply via I2C.

        :param pid_supply: PID supply as an 8-bit unsigned integer value.
        :type pid_supply: int
        """
        try:
            self._write_8_bit(i2c_register=PID_SUPPLY, data=pid_supply)
        except Exception as e:
            print(f"Failed to write PID supply: {e}")
    def get_alarm_latch_clear(self) -> int:
        """
        Reads the Alarm Latch Clear via I2C.

        :returns: Alarm latch clear state as an 8-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(registeraddress=ALARM_LATCH_CLEAR)
        except Exception as e:
            print(f"Failed to read software reset state: {e}")
            return -1
    def set_alarm_latch_clear(self, alarm_latch_clear: int = 1) -> None:
        """
        Writes the Alarm Latch Clear via I2C.
        """
        try:
            self._write_8_bit(i2c_register=ALARM_LATCH_CLEAR, data=alarm_latch_clear)
        except Exception as e:
            print(f"Failed to write alarm latch clear: {e}")
    def get_operation_mode(self) -> int:
        """
        Reads the Operation Mode via I2C.

        :returns: Operation mode as an 8-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=OPERATION_MODE)
        except Exception as e:
            print(f"Failed to read operation mode: {e}")
            return -1
    def set_operation_mode(self, operation_mode: int) -> None:
        """
        Writes the Operation Mode via I2C.

        :param operation_mode: Operation mode as an 8-bit unsigned integer value.
        :type operation_mode: int
        """
        try:
            self._write_8_bit(i2c_register=OPERATION_MODE, data=operation_mode)
        except Exception as e:
            print(f"Failed to write operation mode: {e}")
    def get_relative_zero_value(self) -> float:
        """
        Reads the Relative Zero Value via I2C.

        :returns: Relative zero value as a 32-bit float value.
        :rtype: float
        """
        try:
            return self._read_32_float(i2c_register=NEW_RELATIVE_ZERO)
        except Exception as e:
            print(f"Failed to read relative zero value: {e}")
            return -1.0
    def set_relative_zero_value(self, relative_zero_value: float) -> None:
        """
        Writes the Relative Zero Value via I2C.

        :param relative_zero_value: Relative zero value as a 32-bit float value.
        :type relative_zero_value: float
        """
        try:
            self._write_32_float(i2c_register=NEW_RELATIVE_ZERO, data=relative_zero_value)
        except Exception as e:
            print(f"Failed to write relative zero value: {e}")
    def get_relative_zero_state(self) -> int:
        """
        Reads the Relative Zero State via I2C.

        :returns: Relative zero state as an 8-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(i2c_register=USE_NEW_RELATIVE_ZERO)
        except Exception as e:
            print(f"Failed to read relative zero state: {e}")
            return -1
    def set_relative_zero_state(self, relative_zero_state: int) -> None:
        """
        Writes the Relative Zero State via I2C.

        :param relative_zero_state: Relative zero state as an 8-bit unsigned integer value.
        :type relative_zero_state: int
        """
        try:
            self._write_8_bit(i2c_register=USE_NEW_RELATIVE_ZERO, data=relative_zero_state)
        except Exception as e:
            print(f"Failed to write relative zero state: {e}")
    def get_software_reset(self) -> int:
        """
        Reads the Software Reset State via I2C.

        :returns: Software reset state as an 8-bit unsigned integer value.
        :rtype: int
        """
        try:
            return self._read_8_bit(registeraddress=SOFTWARE_RESET)
        except Exception as e:
            print(f"Failed to read software reset state: {e}")
            return -1
    def set_software_reset(self, software_reset: int = 1) -> None:
        """
        Writes the Software Reset to registers.

        :param software_reset: Software reset state as an 8-bit unsigned integer value.
        :type software_reset: int
        """
        try:
            self._write_8_bit(i2c_register=SOFTWARE_RESET, data=software_reset)
        except Exception as e:
            print(f"Failed to write software reset: {e}")

class UARTContinuous:
    """Base functionality for UART continuous output Protocol

    **Quickstart: Importing and using Ion_Connect UART continuous output Protocol**

            Here is an example of using the :class:`UARTContinuous`.
            First you will need to import the libraries to use the sensor

            .. code-block:: python

                import ION_Connect

            Once this is done you can define your `Ion_Connect.UARTContinuous` object and define the object.

            .. code-block:: python

                port = '/dev/ttyUSB0'
                baudrate = 9600
                timeout = 1
                ion_uart_cont = Ion_Connect.UARTContinuous(port, baudrate, timeout)

            Now you have access to the registers.
            The following example will read the sensor mV and return a float to :variable:'sensor_mv'.

            .. code-block:: python

                sensor_mv = ion_uart_cont.get_sensor_mv()

    """    
    def __init__(self,  port: str, baudrate: int = 9600, timeout: int = 1, parity: str = 'E', stopbits: int = 1, bytesize: int = 8) -> None:
        """
        Initialisation of the UART continuous output protocol.

        :param port: Port to query from.
        :type port: str
        :param baudrate: Baudrate used for communication.
        :type baudrate: int
        :param timeout: Timeout after queries sent.
        :type timeout: int
        """
        if not port:
            raise ConnectionError("No port has been defined!")
        self._port = port
        self._baudrate = baudrate
        self._timeout = timeout
        self._parity = parity
        self._stopbits = stopbits
        self._bytesize = bytesize
        self._uart = None
    @property
    def port(self) -> str:
        """
        Gets the port.
                    
        :returns: port value.
        :rtype: str
        """
        return self._port
    @property
    def baudrate(self) -> int:
        """
        Gets the baudrate            
            
        :returns: baudrate value.
        :rtype: int
        """
        return self._baudrate
    @property
    def timeout(self) -> int:
        """
        Gets the timeout.
            
        :returns: timeout value.
        :rtype: int
        """
        return self._timeout
    @property
    def parity(self) -> str:
        """
        Gets the parity.
            
        :returns: parity value.
        :rtype: int
        """
        return self._parity
    @property
    def stopbits(self) -> int:
        """
        Gets the stopbits.
            
        :returns: stopbits value.
        :rtype: int
        """
        return self._stopbits
    @property
    def bytesize(self) -> int:
        """
        Gets the bytesize.
            
        :returns: bytesize value.
        :rtype: int
        """
        return self._bytesize
    def open(self) -> None:
        """Opens the UART connection."""
        if self._uart is None:
            try:
                self._uart = Serial(
                    port = self._port, 
                    baudrate = self._baudrate, 
                    timeout = self._timeout, 
                    parity = self._parity, 
                    stopbits = self._stopbits, 
                    bytesize = self._bytesize,
                )
                self._uart.flush()
            except Exception as e:
                raise ConnectionError(f"Failed to open UART connection: {e}")
    def close(self) -> None:
        """Closes the UART connection."""
        if self._uart is not None:
            try:
                self._uart.close()
            except Exception as e:
                print(f"Failed to close UART connection: {e}")
            finally:
                self._uart = None
    def __del__(self) -> None:
        """Ensures the UART connection is closed when the object is deleted."""
        self.close()
    def flush(self) -> None:
        """Flushes the UART serial buffer."""
        self._uart.reset_input_buffer()
        self._uart.reset_output_buffer()
    @staticmethod
    def _bytes_to_float32(byte_data: bytes) -> float:
        """
        Converts a sequence of bytes to a 32-bit floating-point number.

        :param byte_data: A bytes object containing exactly 4 bytes.
        :type byte_data: bytes
        :returns: A 32-bit floating-point number derived from the byte data.
        :rtype: float
        :raises ValueError: If the input does not contain exactly 4 bytes.
        """
        if len(byte_data) != 4:
            raise ValueError("Input must be exactly 4 bytes")

        int_value = int.from_bytes(byte_data, byteorder='big', signed=False)
        sign = (int_value >> 31) & 0x01
        exponent = (int_value >> 23) & 0xFF
        mantissa = int_value & 0x7FFFFF

        if exponent == 0:
            float32_value = (-1)**sign * 2**(-126) * (mantissa / 2**23)
        elif exponent == 0xFF:
            if mantissa == 0:
                float32_value = float('inf') if sign == 0 else float('-inf')
            else:
                float32_value = float('nan')
        else:
            float32_value = (-1)**sign * 2**(exponent - 127) * (1 + mantissa / 2**23)

        return float32_value    
    @staticmethod
    def _calculate_crc16(data: bytes) -> int:
        """
        Calculates the CRC16.

        :param data: Data bytes received.
        :type data: bytes
        :returns: CRC16 value.
        :rtype: int
        """
        crc = 0xFFFF
        for byte in data:
            crc ^= byte
            for _ in range(8):
                if crc & 0x0001:
                    crc >>= 1
                    crc ^= 0xA001
                else:
                    crc >>= 1
        return crc
    def _check_checksum(self, buffer: bytes) -> bool:
        """
        Checks the CRC16 of the UART continuous output.

        :param buffer: Checksum data input.
        :type buffer: bytes
        :returns: Boolean of CRC16 calculation.
        :rtype: bool
        """
        calculated_crc = self._calculate_crc16(buffer[:29])
        actual_crc = int.from_bytes(buffer[29:31], byteorder='little')
        return actual_crc == calculated_crc
    def _read_uart(self) -> bytes:
        """
        Reads data from UART and checks checksum.

        :returns: Buffer data if checksum is valid.
        :rtype: bytes
        :raises ValueError: If checksum is invalid.
        """
        buffer = self._uart.read(size=31)
        if self._check_checksum(buffer):
            return buffer
        else:
            raise ValueError("Checksum Error")

    def get_device_address(self) -> int:
        """
        Reads the Device Address from UART continuous output.

        :returns: Device address as an integer value.
        :rtype: int
        """
        buffer = self._read_uart()
        return buffer[0]
    def get_function_code(self) -> int:
        """
        Reads the Function Code from UART continuous output.

        :returns: Function code as an integer value.
        :rtype: int
        """
        buffer = self._read_uart()
        return buffer[1]
    def get_length(self) -> int:
        """
        Reads the Length from UART continuous output.

        :returns: Length as an integer value.
        :rtype: int
        """
        buffer = self._read_uart()
        return buffer[2]
    def get_sensor_ppm(self) -> float:
        """
        Reads the Sensor PPM from UART continuous output.

        :returns: Sensor PPM as a float value.
        :rtype: float
        """
        buffer = self._read_uart()
        return self._bytes_to_float32(buffer[3:7])
    def get_sensor_mv(self) -> float:
        """
        Reads the Sensor MV from UART continuous output.

        :returns: Sensor MV as a float value.
        :rtype: float
        """
        buffer = self._read_uart()
        return round(self._bytes_to_float32(buffer[7:11]), 4)
    def get_dac_value(self) -> float:
        """
        Reads the DAC Value from UART continuous output.

        :returns: DAC value as a float value.
        :rtype: float
        """
        buffer = self._read_uart()
        return self._bytes_to_float32(buffer[11:15])
    def get_sensor_error_status(self) -> int:
        """
        Reads the Sensor Error Status from UART continuous output.

        :returns: Sensor error status as aN unsigned integer value.
        :rtype: float
        """
        buffer = self._read_uart()
        return buffer[15]
    def get_sensor_type(self) -> int:
        """
        Reads the Sensor Type from UART continuous output.

        :returns: Sensor type as aN unsigned integer value.
        :rtype: int
        """
        buffer = self._read_uart()
        return buffer[16]
    def get_temperature(self) -> float:
        """
        Reads the Temperature from UART continuous output.

        :returns: Temperature value as a float value.
        :rtype: float
        """
        buffer = self._read_uart()
        return self._bytes_to_float32(buffer[17:21])
    def get_humidity(self) -> float:
        """
        Reads the Humidity from UART continuous output.

        :returns: Humidity as a float value.
        :rtype: float
        """
        buffer = self._read_uart()
        return self._bytes_to_float32(buffer[21:25])   
    def get_pressure(self) -> float:
        """
        Reads the Pressure from UART continuous output.

        :returns: Pressure as a float value.
        :rtype: float
        """
        buffer = self._read_uart()
        return self._bytes_to_float32(buffer[25:29])
    def get_checksum_hi(self) -> int:
        """
        Reads the Checksum CRC Hi from UART continuous output.

        :returns: Checksum CRC Hi as an integer value.
        :rtype: int
        """
        buffer = self._read_uart()
        return buffer[29]
    def get_checksum_lo(self) -> int:
        """
        Reads the Checksum CRC Lo from UART continuous output.

        :returns: Checksum CRC Lo as an integer value.
        :rtype: int
        """
        buffer = self._read_uart()
        return buffer[30]   
  
    def test_uart(self) -> None:
        """
        Reads data from UART and prints the buffer.

        :raises ValueError: If the UART connection is not open.
        """
        if self._uart is None:
            raise ValueError("UART connection is not open.")
        
        try:
            buffer = self._uart.read(size=31)
            print(f"{buffer[0]}")
            print(f"{buffer[1]}")
            print(f"{buffer[2]}")
            print(f"{buffer[3]}")
            print(f"{buffer[4]}")
            print(f"{buffer[5]}")
            print(f"{buffer[6]}")
            print(f"{buffer[7]}")
            print(f"{buffer[8]}")
            print(f"{buffer[9]}")
            print(f"{buffer[10]}")
            print(f"{buffer[11]}")
            print(f"{buffer[12]}")
            print(f"{buffer[13]}")
            print(f"{buffer[14]}")
            print(f"{buffer[15]}")
            print(f"{buffer[16]}")
            print(f"{buffer[17]}")
            print(f"{buffer[18]}")
            print(f"{buffer[19]}")
            print(f"{buffer[20]}")
            print(f"{buffer[21]}")
            print(f"{buffer[22]}")
            print(f"{buffer[23]}")
            print(f"{buffer[24]}")
            print(f"{buffer[25]}")
            print(f"{buffer[26]}")
            print(f"{buffer[27]}")
            print(f"{buffer[28]}")
            print(f"{buffer[29]}")
            print(f"{buffer[30]}")
        except Exception as e:
            print(f"Failed to read from UART: {e}")
